-- Calculation of HMO  by  H. KIHARA
with Ada.Text_IO; use Ada.Text_IO;
with Ada.Numerics.Elementary_Functions;
use Ada.Numerics.Elementary_Functions;

procedure HMO is 

  package INT_IO is new INTEGER_IO(INTEGER);
  package REAL_IO is new FLOAT_IO(FLOAT);
  use INT_IO, REAL_IO;

  type   Matrix is array(1..30, 1..30) of FLOAT ;
    h : Matrix ;        -- secular determinat
    v : Matrix ;        -- coefficient matrix
    n : integer ;       -- number of atoms
    np : integer ;       -- number of electrons
	
  procedure Input(h: out Matrix ; natom: out integer; np: out integer ) is

   compound : string(1 .. 40) ;    -- compound name
   nbond    : integer ;       -- number of bonds
   nhetero  : integer ;       -- number of matrix elements to be changed
   hij      : FLOAT ;
   ia, ib : integer ;
   L: natural;
   
   begin
   PUT_LINE("  HMO Calculation Program") ;
   PUT("Name of Compound  : ") ;
   GET_LINE(compound, L) ; NEW_LINE;
   PUT("Number of atoms: ");
   GET(natom) ; NEW_LINE;
   PUT("Number of bonds: ");
   GET(nbond) ; NEW_LINE;
   PUT("Number of electrons: ");
   GET(np) ; NEW_LINE;
   PUT("Number of matrix elements to be changed: ");
   GET(nhetero) ; NEW_LINE;

   for i in 1..natom loop
      for j in 1.. natom loop
         h(i, j) := 0.0 ;
      end loop;
   end loop;

   PUT_LINE("Enter the number pair (i, j) of the atoms bonded") ;
   for k in 1 .. nbond loop
      PUT("  Bond ") ;   PUT(k) ; PUT(": ") ; 
      GET(ia) ;
	  GET(ib) ; NEW_LINE;
      h(ia,ib) := 1.0 ;
      h(ib,ia) := 1.0 ;
   end loop;

   if nhetero > 0 then
      PUT_LINE("Enter the matrix element to change (i, j, Hij)");
      for k in 1 .. nhetero loop
        PUT(k) ; PUT(" : ") ;
        GET(ia) ;
        GET(ib) ;
        GET(hij) ; NEW_LINE;
        h(ia,ib) := hij ;
        h(ib,ia) := hij ;
      end loop;
   end if;
 end Input;

-- Diagonalization of matrix by Jacobi method
  procedure  Jacobi(h: in out Matrix ; v: out Matrix ; n: in integer) is

	threshold: constant FLOAT := 0.00001 ;  
    itmax: constant INTEGER := 100 ;   -- maximaum iteration cycles

   d,sm,hj,hk,ta,tb,sn,cs,hkl,hjl,vj,vk : FLOAT   ;
     it : integer ;

  begin
   it := 1 ;
   d := 1.0 ;
 	 -- unit matrix
   for i in 1 .. n loop
      for j in  1 .. n loop
         v(i,j) := 0.0 ;
      end loop;
    end loop;		 
   for i in 1 ..n loop
     v(i,i) := 1.0 ; 
    end loop;
	
   while (d > threshold) and (it < itmax) loop     -- Convergence judgment
      d := abs(h(2,1)) ;
      for i in 2 .. n loop
         for j in 1 .. i-1 loop
  		   if abs(h(i,j)) > d then d := abs(h(i,j)); end if;
         end loop; -- of for j
     end loop; -- of for i 
	  
      if d > threshold then

         sm := 0.1*d ;
         it := it + 1 ;
         for j in 2 .. n loop

 		   for k in 1 .. j-1 loop

               if abs(h(j,k)) > sm then

                  hj := h(j,j) ;
                  hk := h(k,k) ;
                  tb := (hj-hk)/(2.0*h(k,j)) ;
				  
                  if tb <= 0.0 then
  				    ta := tb - sqrt(1.0+tb*tb);
                  else
				    ta := tb + sqrt(1.0+tb*tb) ;
				  end if;
				  
                  cs := 1.0/sqrt(1.0+ta*ta) ;
                  sn := cs*ta ;
				  
                  for l in 1 .. n loop

                     if l = k then
                        h(k,k) := hk*cs*cs+h(j,k)*2.0*sn*cs+hj*sn*sn ;
                        h(j,j) := hk+hj-h(k,k) ;					 
                     else if l /= j then
                          hkl := h(k,l) ;
                          hjl := h(j,l) ;
                          h(k,l) := hkl*cs+hjl*sn ;
                          h(j,l) := hjl*cs-hkl*sn ;
                          h(l,k) := h(k,l) ;
                          h(l,j) := h(j,l);
                        end if;
                     end if;
					 
                  end loop; 
				  
                  h(j,k) := h(j,k)*(cs*cs-sn*sn)+(hj-hk)*sn*cs ;
                  h(k,j) := h(j,k) ;
				  
                  for i in 1 .. n loop
                     vj := v(i,j) ;
                     vk := v(i,k) ;
                     v(i,j) := vj*cs-vk*sn ;
                     v(i,k) := vj*sn+vk*cs ;
                  end loop;
				  
               end if; 
			   
            end loop;  -- of for k 
         end loop; -- of for j 
      end if;
	  
   end loop;
end ; -- { of Jacobi }

  procedure SWAP_VAL(a,b: in out FLOAT ) is
  temp : FLOAT ;
    begin
	  temp := a ;
	  a := b ;
	  b := temp ;
  end SWAP_VAL;

-- Sort eigenvalues and eigenvectors in ascending order of eigenvalues
  procedure SORT(h,v : in out Matrix ; n : in integer) is

  max : integer ;
  emax : FLOAT   ;

  begin
    for i in 1..n loop
      emax := h(i,i) ;
      max := i ;
	  
      for j in i..n loop
         if h(j,j) > emax then
            max := j ;
            emax := h(j,j) ;
         end if;
      end loop;
	  
      if i< max then
         SWAP_VAL(h(i,i),h(max,max)) ;
		 
         for j in 1 .. n loop
		   SWAP_VAL(v(j,i),v(j,max)) ;
		 end loop;
		 
      end if;
    end loop;
   
  end SORT;

-- Print eigenvalues and eigenvectors
  procedure Print_Eig(h,v : in Matrix ; n : in integer) is

  first, last: integer ;
  column : constant integer := 8 ;     -- number of items printed in a aline 
  
  begin
   first := 1 ;
   last := first + column - 1 ;
   NEW_LINE; 
   PUT("--  Eigen values and Eigen vector   --") ;
    
   while first <= n loop
     NEW_LINE;
     if last > n then last := n ; end if;
      PUT("  ");
      for i in  first..last loop
	    PUT(h(i,i),4,5,0) ;
	  end loop;
      NEW_LINE;
	  
      for k in  1..n loop
         PUT(k, 2) ;
         for j in first..last loop
   		   PUT(v(k,j),4,5,0) ;
 	     end loop;
         NEW_LINE;
     end loop;
     NEW_LINE;

     first := last + 1 ;
     last := first + column - 1 ;
   end loop;

  end Print_Eig;

   -- 結合次数の計算と表示
  Procedure PMAT(h, v : in Matrix; n: in Integer; np: in Integer) is

    P : Matrix ;        -- bondorder matrix
    m, r: Integer;
	NOCC: Integer;
    Odd: Boolean := False;
    Degenerate: Boolean := False;
    EPS: constant Float := 0.000001;
    nItem : constant Integer := 8;

	begin
       NOCC := (np + 1) / 2; 
       If NOCC * 2 > np Then Odd := True; end if;

    -- HOMOと一つ上のMOのエネルギーを比較
       If h(NOCC, NOCC) - h(NOCC + 1, NOCC + 1) < EPS Then Degenerate := True; end if;

       For i in 1 .. n loop
          For j in  1 .. n loop
             P(i, j) := 0.0;
           end loop;
       end loop; 

       For i in 1 .. N loop

          For j in  1 .. i loop

             For k in 1 .. NOCC - 1 loop
                 P(i, j) := P(i, j) + 2.0 * V(i, k) * V(j, k);
             end loop;

             If Degenerate Then
                 If Odd Then
                     P(i, j) := P(i, j) + 0.5 * (V(i, NOCC) * V(j, NOCC) + V(i, NOCC + 1) * V(j, NOCC + 1));
                 Else
                     P(i, j) := P(i, j) + V(i, NOCC) * V(j, NOCC) + V(i, NOCC + 1) * V(j, NOCC + 1);
                 End If;
             Else
                 If Odd Then
                     P(i, j) := P(i, j) + V(i, NOCC) * V(j, NOCC);
                 Else
                     P(i, j) := P(i, j) + 2.0 * V(i, NOCC) * V(j, NOCC);
                 End If;

             End If;
			 
			 if i /= j then P(j, i) := P(i, j); end if;
			 
          end loop;  -- { for j}

      end loop;  -- { for i}

    -- 結合次数の表示
     Put_Line("-- Bond Order Matrix --");
     r := 1;
     m := r + nItem - 1;

     While r <= n loop

        If m > n Then m := n; end if;
        put("  ");
        For i in  r .. m loop
           put(i, 9);
        end loop;
        NEW_LINE;

        For i in  1 .. n loop
           put(i, 4);
           For j in  r .. m loop
              Put( P(i, j), 3,5,0);
           end loop;
           NEW_LINE;
        end loop;

        r := m + 1;
        m := r + nItem - 1;
        NEW_LINE;

      end loop;

   End PMAT;
	
  begin

   Input(h, n, np) ;
   
   Jacobi(h, v, n) ;

   Sort(h, v, n) ;

   Print_Eig(h, v, n) ;

   PMAT(h, v, n, np) ;

end HMO;
